<?php

/**
 * @file
 * Endpoint functions for the User Access Hub module.
 */

/**
 * Handle handshake.
 */
function useraccesshub_endpoint_handshake() {

  // Only proceed if handshakes are enabled.
  if (!variable_get('useraccesshub_handshake_enabled')) {
    return _useraccesshub_error_response_handshake();
  }

  if (!_useraccesshub_validate_method(['POST'])) {
    return _useraccesshub_error_response_method();
  }

  if (!_useraccesshub_validate_api_key()) {
    return _useraccesshub_error_response_key();
  }

  $body = file_get_contents('php://input');
  if ($errors = _useraccesshub_validate_body($body, array('public_key', 'site_id'))) {
    return _useraccesshub_error_response_body($errors);
  }

  // Set config values.
  $body = json_decode($body);
  variable_set('useraccesshub_public_key', $body->public_key);
  variable_set('useraccesshub_site_id', $body->site_id);
  variable_set('useraccesshub_handshake_enabled', FALSE);

  $output = [
    'roles' => user_roles(TRUE),
  ];

  return json_encode($output);
}

/**
 * Handle roles.
 */
function useraccesshub_endpoint_roles() {
  if (!_useraccesshub_validate_method(['POST'])) {
    return _useraccesshub_error_response_method();
  }

  if (!$_SERVER['HTTP_SIGNATURE']) {
    return _useraccesshub_error_response_signature();
  }

  $body = file_get_contents('php://input');
  $signature = base64_decode($_SERVER['HTTP_SIGNATURE']);
  if (!$signature || !_useraccesshub_validate_signature($body, $signature)) {
    return _useraccesshub_error_response_signature();
  }

  if ($errors = _useraccesshub_validate_body($body, array('created', 'expiry'))) {
    return _useraccesshub_error_response_body($errors);
  }

  $message = json_decode($body);
  if (!_useraccesshub_validate_times($message)) {
    return _useraccesshub_error_response_times();
  }

  $output = [
    'roles' => user_roles(TRUE),
  ];

  return json_encode($output);
}

/**
 * Handle login.
 */
function useraccesshub_endpoint_sso() {

  // Only proceed with the login if the module functionality is enabled.
  if (!variable_get('useraccesshub_enabled')) {
    return _useraccesshub_error_response_login();
  }

  if (!_useraccesshub_validate_method(['GET', 'POST'])) {
    return _useraccesshub_error_response_method();
  }

  switch ($_SERVER['REQUEST_METHOD']) {

    // The initial request will hit this URL, which will redirect immediately
    // to the User Access Hub for user login/validation.
    case 'GET':
      drupal_add_http_header('Status', '302 Found');
      drupal_add_http_header('Location', 'https://www.useraccesshub.com.com/sso/request?site=' . variable_get('useraccesshub_site_id'));
      break;

    // Once the user has logged into the User Access Hub and is validated, the
    // hub will post a signed message to the site, signalling that the user is
    // good to go.
    case 'POST':

      // Message is required.
      $message = $_POST['message'];
      if (!$message) {
        return _useraccesshub_error_response_message();
      }

      // Message needs to be base64 decoded.
      $message = base64_decode($message);
      if (!$message) {
        return _useraccesshub_error_response_message();
      }

      if ($errors = _useraccesshub_validate_body($message, array('data', 'signature'))) {
        return _useraccesshub_error_response_body($errors);
      }

      $message = json_decode($message);
      $data = json_encode($message->data);
      $properties = array('email', 'roles', 'created', 'expiry');
      if ($errors = _useraccesshub_validate_body($data, $properties)) {
        return _useraccesshub_error_response_body($errors);
      }

      $signature = base64_decode($message->signature);
      if (!$signature || !_useraccesshub_validate_signature($data, $signature)) {
        return _useraccesshub_error_response_signature();
      }

      if (!_useraccesshub_validate_times($message->data)) {
        return _useraccesshub_error_response_times();
      }

      // Load or create the user.
      $account = user_load_by_mail($message->data->email);
      if ($account) {

        // Make sure the user is active.
        if (!$account->status) {
          return _useraccesshub_error_response_blocked();
        }

        // If local accounts are allowed, don't change anything about the
        // user (except adding the role below).
        if (!variable_get('useraccesshub_allow_local')) {
          $account->roles = array();
          $account->pass = _useraccesshub_generate_key(128);
        }

        $account = user_save($account);
      }
      else {
        $account = array(
          'name' => $message->data->email,
          'mail' => $message->data->email,
          'pass' => _useraccesshub_generate_key(128),
          'status' => 1,
        );

        $account = user_save(NULL, $account);
      }

      // Set the roles.
      foreach ($message->data->roles as $rid) {
        if ($rid == 0) {
          user_multiple_role_edit(array($account->uid), 'add_role', variable_get('useraccesshub_default_role'));
        }
        else {
          user_multiple_role_edit(array($account->uid), 'add_role', $rid);
        }
      }

      // Log the user in.
      global $user;
      $user = $account;
      user_login_finalize();

      // Redirect.
      drupal_add_http_header('Status', '302 Found');
      drupal_add_http_header('Location', variable_get('useraccesshub_redirect', '/user'));
      break;
  }

  return '';
}

/**
 * Generic error response.
 *
 * @param array $error
 *   The body of the error to return.
 * @param int $http_code
 *   The status code to return.
 *
 * @return string
 *   The error response.
 */
function _useraccesshub_error_response($error, $http_code) {
  switch ($http_code) {
    case '400':
      drupal_add_http_header('Status', '400 Bad Request');
      break;

    case '401':
      drupal_add_http_header('Status', '401 Unauthorized');
      break;

    case '403':
      drupal_add_http_header('Status', '403 Forbidden');
      break;

    case '405':
      drupal_add_http_header('Status', '405 Method Not Allowed');
      break;
  }

  return json_encode($error);
}

/**
 * Blocked account error response.
 *
 * @return string
 *   The error response.
 */
function _useraccesshub_error_response_blocked() {
  $message = array('message' => 'Unable to login the user since they are blocked on the site.');
  return _useraccesshub_error_response($message, 403);
}

/**
 * Error response if the body is malformed.
 *
 * @param string[] $errors
 *   The list of errors to add to the response.
 *
 * @return string
 *   The error response.
 */
function _useraccesshub_error_response_body($errors) {
  $message = array(
    'message' => 'The body of the request is required.',
    'errors' => $errors,
  );
  return _useraccesshub_error_response($message, 400);
}

/**
 * Handshake error response.
 *
 * @return string
 *   The error response.
 */
function _useraccesshub_error_response_handshake() {
  $message = array(
    'error' => 1,
    'message' => 'Handshake is forbidden.',
  );
  return _useraccesshub_error_response($message, 200);
}

/**
 * Login error response.
 *
 * @return string
 *   The error response.
 */
function _useraccesshub_error_response_login() {
  $message = array('message' => 'Login functionality using this method is disabled.');
  return _useraccesshub_error_response($message, 403);
}

/**
 * API Key error response.
 *
 * @return string
 *   The error response.
 */
function _useraccesshub_error_response_key() {
  $message = array('message' => 'The API key is invalid.');
  return _useraccesshub_error_response($message, 401);
}

/**
 * Message error response.
 *
 * @return string
 *   The error response.
 */
function _useraccesshub_error_response_message() {
  $message = array('message' => 'The message POST variable is required and needs to be base 64 encoded.');
  return _useraccesshub_error_response($message, 405);
}

/**
 * HTTP method error response.
 *
 * @return string
 *   The error response.
 */
function _useraccesshub_error_response_method() {
  $message = array('message' => 'The method is not allowed.');
  return _useraccesshub_error_response($message, 405);
}

/**
 * Signature error response.
 *
 * @return string
 *   The error response.
 */
function _useraccesshub_error_response_signature() {
  $message = array('message' => 'The signature is invalid.');
  return _useraccesshub_error_response($message, 401);
}

/**
 * Time error response.
 *
 * @return string
 *   The error response.
 */
function _useraccesshub_error_response_times() {
  $message = array('message' => 'The login request has expired.');
  return _useraccesshub_error_response($message, 401);
}

/**
 * Validate API key.
 *
 * @return bool
 *   TRUE if the apikey header value matches key and FALSE otherwise.
 */
function _useraccesshub_validate_api_key() {
  if (isset($_SERVER['HTTP_APIKEY']) && $_SERVER['HTTP_APIKEY'] == variable_get('useraccesshub_api_key')) {
    return TRUE;
  }

  return FALSE;
}

/**
 * Validate request body.
 *
 * @param string $body
 *   The body of the request.
 * @param array $properties
 *   The properties required for the body.
 *
 * @return string[]
 *   The array of errors.
 */
function _useraccesshub_validate_body($body, $properties) {
  $errors = [];

  if (empty($body)) {
    $errors[] = 'Body is missing and is required.';
  }
  else {
    $json_body = @json_decode($body);
    if (json_last_error() != JSON_ERROR_NONE) {
      $errors[] = 'Invalid JSON: ' . json_last_error_msg();
    }

    foreach ($properties as $property) {
      if (!isset($json_body->{$property})) {
        $errors[] = "The '" . $property . "' property is missing";
      }
    }
  }

  return $errors;
}

/**
 * Validate HTTP method.
 *
 * @param array $methods
 *   The accepted methods.
 *
 * @return bool
 *   TRUE if the http method matches the method's passed in and FALSE
 *   otherwise.
 */
function _useraccesshub_validate_method($methods) {
  if (in_array($_SERVER['REQUEST_METHOD'], $methods)) {
    return TRUE;
  }

  return FALSE;
}

/**
 * Validate Signature.
 *
 * @param string $data
 *   The data to validate.
 * @param string $signature
 *   The signature to validate.
 *
 * @return bool
 *   TRUE if the signature header value matches key and FALSE otherwise.
 */
function _useraccesshub_validate_signature($data, $signature) {
  if (openssl_verify($data, $signature, variable_get('useraccesshub_public_key'), OPENSSL_ALGO_SHA384) == 1) {
    return TRUE;
  }

  return FALSE;
}

/**
 * Validate times.
 *
 * @param object $data
 *   The message data.
 *
 * @return bool
 *   TRUE if the times are in the correct range and FALSE otherwise.
 */
function _useraccesshub_validate_times($data) {

  // These dates will be passed in ISO 8601 date format.
  $created = strtotime($data->created);
  $expiry = strtotime($data->expiry);

  // The request from the User Access Hub must have been created before now
  // and must expire after now.
  $now = time();
  if ($created <= $now && $expiry >= $now) {
    return TRUE;
  }

  return FALSE;
}
